/*
================================================================================
 File: fat32.h
--------------------------------------------------------------------------------
 Author:        Piotr Wojtowicz
 Purpose:       FAT16/32 read-only library MSM compatible
 Compiler:      ANSI C
 Hardware:      
 Doc:
--------------------------------------------------------------------------------
 Includes:      
--------------------------------------------------------------------------------
 Comments:                      UWAGA!!!
                Oprogramowanie to udostępnione jest do celów poglądowych,
                nie jest przetestowane pod kątem bezpieczeństwa ani spełnienia
                jakichkolwiek norm wymaganych aby urządzenie używające go
                mogło być dopuszczone do użytku, dlatego używasz go
                na własną odpowiedzialność a jego autor nie ponosi żadnego
                typu odpowiedzialności za jakiekolwiek skutki jego użycia.
                Jeżeli nie zgadzasz się z powyższym warunkiem proszę
                o nieużywanie tego oprogramowania.
================================================================================
*/

#include "fat32.h"
#include <string.h>


static MSMU_Desc FAT_User;
static FAT_Desc FAT_Devices[FAT_MAX_NUM_DEVICES];
static FAT_UserDesc *FAT_Users[FAT_MAX_NUM_USERS];
static uint32_t bNumFATUsers = 0;
static uint32_t bNumFATDevices = 0;

static union {
	uint8_t uint8[512/sizeof(uint8_t)];
	uint16_t uint16[512/sizeof(uint16_t)];
	uint32_t uint32[512/sizeof(uint32_t)];
}FAT_Cache;
//static uint8_t FAT_Cache[512];
static uint32_t FAT_Cache_stored_sector = 0xFFFFFFFF;
static uint32_t FAT_Cache_owner = FAT_MAX_NUM_DEVICES + 1;
static uint8_t FAT_LFN_Buffer[260];


#if(OS_SUPPORT == 1)

#else
#define FAT_CACHE_SEMAPHORE_GET()
#define FAT_CACHE_SEMAPHORE_RETURN()
#define FAT_CACHE_SEMAPHORE_INIT()

#define FAT_LFN_SEMAPHORE_GET()
#define FAT_LFN_SEMAPHORE_RETURN()
#define FAT_LFN_SEMAPHORE_INIT()
#endif



static void FAT_ConnectEvent (MSMD_Desc *MSD);



static uint32_t FAT_GetAllocTableRecord(FAT_Desc *FAT, uint32_t Cluster) {
	uint32_t alloc_table_sector;
	uint32_t alloc_table_sector_offset;
	uint64_t address;
	uint32_t record = 0;

	if(FAT->FAT_Type == 16) {
		Cluster <<= 1;
	}
	else if(FAT->FAT_Type == 32) {
		Cluster <<= 2;
	}

	// here we calculate number of sector that stores requied
	// cluster record in allocation table
	alloc_table_sector = Cluster / FAT->Param.BPB_BytsPerSec;//FAT->Param.BPB_RsvdSecCnt
//		+ (Cluster / FAT->Param.BPB_BytsPerSec);
	alloc_table_sector += FAT->Param.BPB_RsvdSecCnt;

	// here we calculate in wich 32-bit word of alloc_table_sector
	// requied record is
	alloc_table_sector_offset = Cluster % FAT->Param.BPB_BytsPerSec;

	FAT_CACHE_SEMAPHORE_GET();

	// if data is not stored in FAT cache we need to read this sector
	if((alloc_table_sector != FAT_Cache_stored_sector)
			|| FAT->FAT_ID != FAT_Cache_owner) {
		address = ((uint64_t)alloc_table_sector) * ((uint64_t)FAT->Param.BPB_BytsPerSec);
		FAT_Cache_stored_sector = alloc_table_sector;
		FAT_Cache_owner = FAT->FAT_ID;
		
		MSMD_ReadBlock(FAT->MSD, &FAT_User, FAT_Cache.uint8, (uint32_t)address, ((uint32_t)(address >> 32)), 1);
	}

	// here we read record from the buffer and return this data
	if(FAT->FAT_Type == 16) {
		record = (uint32_t)(*((uint16_t*)&(FAT_Cache.uint8[alloc_table_sector_offset])));
	}
	else if(FAT->FAT_Type == 32) {
		record = (*((uint32_t*)&(FAT_Cache.uint8[alloc_table_sector_offset]))) & 0xFFFFFFF;
	}

	FAT_CACHE_SEMAPHORE_RETURN();

	return(record);
}//*/



static uint16_t FAT_File_alloc_table_buffer_fill(FAT_Desc *FAT, FILE *file_desc, uint32_t cluster) {
	uint32_t counter = 1;

	file_desc->AllocTableBuffer[0] = cluster;
	file_desc->AllocTableBufferPosition = 0;

	while(counter < FILE_ALLOC_TABLE_BUFFER_SIZE) {
		cluster = FAT_GetAllocTableRecord(FAT, cluster);
		if(FAT->FAT_Type == 16) {
			if(cluster >= 0xFFF8) {
				file_desc->AllocTableBuffer[counter] = 0;
				file_desc->AllocTableBufferStoredDataSize = counter;
				return(counter);
			}
		}
		else if(FAT->FAT_Type == 32) {
			if(cluster >= 0x0FFFFFF8) {
				file_desc->AllocTableBuffer[counter] = 0;
				file_desc->AllocTableBufferStoredDataSize = counter;
				return(counter);
			}
		}
		file_desc->AllocTableBuffer[counter] = cluster;
		file_desc->AllocTableBufferStoredDataSize = counter;
		counter++;
	}
	return(counter);
}



static int FAT_File_read_next_sector(FILE *file_desc) {
	file_desc->CurrentSector++;
	if(!((file_desc->CurrentSector - file_desc->FAT->FirstDataSector) % file_desc->FAT->Param.BPB_SecPerClus)) {

		file_desc->AllocTableBufferPosition++;

		if(file_desc->AllocTableBufferPosition < file_desc->AllocTableBufferStoredDataSize) {
			if(file_desc->AllocTableBuffer[file_desc->AllocTableBufferPosition]) {
				file_desc->CurrentSector = file_desc->AllocTableBuffer[file_desc->AllocTableBufferPosition];
				file_desc->CurrentSector -= 2;
				file_desc->CurrentSector *= file_desc->FAT->Param.BPB_SecPerClus;
				file_desc->CurrentSector += file_desc->FAT->FirstDataSector;
			}
			else {
				return(0);
			}
		}
		else {
			if(file_desc->AllocTableBufferStoredDataSize == FILE_ALLOC_TABLE_BUFFER_SIZE) {
				FAT_File_alloc_table_buffer_fill(
							file_desc->FAT,
							file_desc,
							file_desc->AllocTableBuffer[FILE_ALLOC_TABLE_BUFFER_SIZE - 1]);
			}
			else {
				return(0);
			}
		}
	}
	if(MSMD_AccessResult_Success == MSMD_ReadBlock(
			file_desc->FAT->MSD,
			&FAT_User,
			(uint8_t*)file_desc->Buffer.uint8,
			file_desc->CurrentSector * file_desc->FAT->Param.BPB_BytsPerSec,
			0,
			1)) {
		file_desc->BufferPosition = 0;
		return(file_desc->CurrentSector);
	}
	return(0);
}



static uint8_t FAT_GetChar(uint8_t unicode_L, uint8_t unicode_H) {
	uint8_t ascii;
	uint16_t unicode = (uint16_t)unicode_L;
	unicode |= ((uint16_t)unicode_H) << 8;
	switch(unicode) {
		case 0x105:
			ascii = 'a';
			break;
		case 0x104:
			ascii = 'A';
			break;
		case 0x107:
			ascii = 'c';
			break;
		case 0x106:
			ascii = 'C';
			break;
		case 0x119:
			ascii = 'e';
			break;
		case 0x118:
			ascii = 'E';
			break;
		case 0x142:
			ascii = 'l';
			break;
		case 0x141:
			ascii = 'L';
			break;
		case 0x144:
			ascii = 'n';
			break;
		case 0x143:
			ascii = 'N';
			break;
		case 0xF3:
			ascii = 'o';
			break;
		case 0xD3:
			ascii = 'O';
			break;
		case 0x15B:
			ascii = 's';
			break;
		case 0x15A:
			ascii = 'S';
			break;
		case 0x17C:
			ascii = 'z';
			break;
		case 0x17A:
			ascii = 'z';
			break;
		case 0x17B:
			ascii = 'Z';
			break;
		case 0x179:
			ascii = 'Z';
			break;
		default:
			ascii = unicode_L;
			break;
	}
	return(ascii);
}



static uint8_t *FAT_LFN_name_read(FAT_DIR_Record *REC, uint8_t *buffer) {
	uint8_t counter = 6;

	buffer--;
	*buffer = FAT_GetChar(REC->LFN.LFN_NameChars3[2], REC->LFN.LFN_NameChars3[3]);
	buffer--;
	*buffer = FAT_GetChar(REC->LFN.LFN_NameChars3[0], REC->LFN.LFN_NameChars3[1]);
	while(counter) {
		counter--;
		buffer--;
		*buffer = FAT_GetChar(REC->LFN.LFN_NameChars2[counter * 2], REC->LFN.LFN_NameChars2[counter * 2 + 1]);
	}
	counter = 5;
	while(counter) {
		counter--;
		buffer--;
		*buffer = FAT_GetChar(REC->LFN.LFN_NameChars1[counter * 2], REC->LFN.LFN_NameChars1[counter * 2 + 1]);
	}
	return(buffer);
}



// returns pointer to next record in buffer or zero if there is no any 
static FAT_DIR_Record *FAT_File_get_next_record(FILE *file_desc, FAT_DIR_Record *REC, uint8_t **file_name) {
	uint8_t FAT_LFN_Checksum[20];
	uint8_t name_counter;
	uint8_t *buffer;
	uint8_t error;
	uint8_t cs;

	REC++;
	// if is not REC inside of "file" buffer
	if((uint8_t*)REC >= (uint8_t*)(file_desc->Buffer.uint8) + 512) {
		FAT_File_read_next_sector(file_desc);
		REC = (FAT_DIR_Record *)file_desc->Buffer.uint8;
	}
	// we have to leave records that's REC->DIR.DIR_Name[0] is equal to 0xE5
	// this indicate that specified record is not valid
	while(REC->DIR.DIR_Name[0] == 0xE5) {
		REC++;
		// if is not REC inside of "file" buffer
		if((uint8_t*)REC >= (uint8_t*)(file_desc->Buffer.uint8) + 512) {
			FAT_File_read_next_sector(file_desc);
			REC = (FAT_DIR_Record *)file_desc->Buffer.uint8;
		}
	}
	// if record's REC->DIR.DIR_Name[0] is equal to 0 then we know, that
	// there are not any other records
	if(!REC->DIR.DIR_Name[0]) {
		return(0);
	}
	// LFN
	if((REC->DIR.DIR_Attr & 0xf) == 0xf) {
		// number of parts of name
		name_counter = REC->LFN.LFN_OrdinalField;
		name_counter &= 0x3f;
		
		// we read name from end to begining - here we read
		// last part of name
		buffer = FAT_LFN_name_read(REC, &FAT_LFN_Buffer[260]);
		FAT_LFN_Checksum[0] = REC->LFN.LFN_Checksum;
		name_counter--;
		if(REC->LFN.LFN_OrdinalField & 0x40) {
			error = 0;
		}
		else {
			error = 1;
		}
		REC++;
		// if is not REC inside of "file" buffer
		if((uint8_t*)REC >= (uint8_t*)(file_desc->Buffer.uint8) + 512) {
			FAT_File_read_next_sector(file_desc);
			REC = (FAT_DIR_Record *)file_desc->Buffer.uint8;
		}

		uint8_t checksum_counter = 1;
		// read next parts of name
		while(name_counter) {
			if(REC->LFN.LFN_OrdinalField
					&& (REC->LFN.LFN_OrdinalField == name_counter)
					&& ((REC->DIR.DIR_Attr & 0xf) == 0xf)
					&& !error) {
				buffer = FAT_LFN_name_read(REC, buffer);
				FAT_LFN_Checksum[checksum_counter] = REC->LFN.LFN_Checksum;
				REC++;
				// if is not REC inside of "file" buffer
				if((uint8_t*)REC >= (uint8_t*)(file_desc->Buffer.uint8) + 512) {
					FAT_File_read_next_sector(file_desc);
					REC = (FAT_DIR_Record *)file_desc->Buffer.uint8;
				}
			}
			else {
				error = 1;
				break;
			}
			checksum_counter++;
			name_counter--;
		}
		if(!error) {
			// checksum
			cs = REC->DIR.DIR_Name[0];
			for(name_counter = 1; name_counter < 11; name_counter++) {
				cs = (((cs << 7) & 0x80) | (cs >> 1));
				cs += REC->DIR.DIR_Name[name_counter];
			}

			// checksum checking
			while(checksum_counter) {
				checksum_counter--;
				if(cs != FAT_LFN_Checksum[checksum_counter]) {
					error = 1;
					break;
				}
			}

			if(!error) {
				if(file_name) {
					*file_name = buffer;
				}
				return(REC);
			}
		}
		else {
			while(!((REC->DIR.DIR_Name[0]) && (REC->DIR.DIR_Name[0] != 0xE5)
					&& ((REC->DIR.DIR_Attr & 0xF) != 0xF))) {
				REC++;
				// if is not REC inside of "file" buffer
				if((uint8_t*)REC >= (uint8_t*)(file_desc->Buffer.uint8) + 512) {
					FAT_File_read_next_sector(file_desc);
					REC = (FAT_DIR_Record *)file_desc->Buffer.uint8;
				}
				
			}
			if(!REC->DIR.DIR_Name[0]) {
				return(0);
			}
		}
	}
	// DIR
	buffer = FAT_LFN_Buffer;
	for(name_counter = 0; name_counter < 11; name_counter++) {
		if(REC->DIR.DIR_Name[name_counter] == ' ') {
			if(name_counter == 7 && !(REC->DIR.DIR_Attr & ATTR_DIRECTORY)) {
				*buffer = '.';
				buffer++;
			}
			continue;
		}
		*buffer = REC->DIR.DIR_Name[name_counter];
		if(name_counter == 7 && !(REC->DIR.DIR_Attr & ATTR_DIRECTORY)) {
			buffer++;
			*buffer = '.';
		}
		buffer++;
	}
	*buffer = 0;
	if(file_name) {
		*file_name = FAT_LFN_Buffer;
	}
	return(REC);
}


// returns 0 when strings are not equal or pointer to next part of path when are equal
static uint8_t *FAT_FILE_Path_compare(uint8_t *path, uint8_t *name) {
	while(*path != '/' && *path) {
		if(((*path >= 'A' && *path <= 'Z') || (*path >= 'a' && *path <= 'z'))
				&& ((*name >= 'A' && *name <= 'Z') || (*name >= 'a' && *name <= 'z'))) {
			if((*path == *name)
					|| ((*path + ('a' - 'A')) == *name)
					|| (*path == (*name + ('a' - 'A')))) {
				path++;
				name++;
				continue;
			}
		}
		else if((*path == *name) && *name) {
			path++;
			name++;
			continue;
		}
		return(0);
	}
	if(*name) {
		return(0);
	}
	path++;
	return(path);
}



static FAT_DIR_Record *FAT_get_file_record_from_dir(FILE *file_desc, uint32_t dir_sector, uint8_t *file_name, uint8_t **path_next_part) {
	uint8_t *parsed_file_name = 0;
	FAT_DIR_Record *REC = (FAT_DIR_Record*)file_desc->Buffer.uint8;
	REC--;
	if(MSMD_AccessResult_Success == MSMD_ReadBlock(file_desc->FAT->MSD, &FAT_User, (uint8_t*)file_desc->Buffer.uint8, dir_sector * file_desc->FAT->Param.BPB_BytsPerSec, 0, 1)) {

		FAT_LFN_SEMAPHORE_GET();

		while(REC) {
			REC = FAT_File_get_next_record(file_desc, REC, &parsed_file_name);
			parsed_file_name = FAT_FILE_Path_compare(file_name, parsed_file_name);
			if(parsed_file_name) {
				*path_next_part = parsed_file_name;
				break;
			}
		}

		FAT_LFN_SEMAPHORE_RETURN();

		if(REC) {
			return(REC);
		}
	}
	return(0);
}



static void FAT_FILE_Desc_init_by_root_dir(FILE *file_desc, FAT_Desc *FAT) {
	file_desc->FAT = FAT;
	file_desc->ParentDirectoryStartedSector = 0;
	file_desc->FileStartedSector = FAT->Root_first_sector;
	file_desc->CurrentSector = FAT->Root_first_sector;
	FAT_File_alloc_table_buffer_fill(FAT, file_desc, FAT->Param.BPB_RootClus);
	file_desc->FileSize = 0;
	file_desc->FilePosition = 0;
	file_desc->Attribute = FILE_TYPE_DIRECTORY;
	file_desc->AccessType = FILE_AccessType_CLOSE;
}



static void FAT_FILE_Desc_init_by_dir(FILE *file_desc, FAT_DIR_Record *REC, uint32_t dir_sector, uint32_t dir_cluster) {
	FAT_Desc *FAT = file_desc->FAT;
//	ParentDirectoryStartedSector have to be set outside of this function
	file_desc->FileStartedSector = dir_sector;
	file_desc->CurrentSector = dir_sector;
	file_desc->FileSize = REC->DIR.DIR_FileSize;
	file_desc->FilePosition = 0;
	FAT_File_alloc_table_buffer_fill(FAT, file_desc, dir_cluster);
	if(REC->DIR.DIR_Attr & ATTR_DIRECTORY) {
		file_desc->Attribute = FILE_TYPE_DIRECTORY;
	}
	file_desc->AccessType = FILE_AccessType_CLOSE;
}



static uint32_t FAT_FILE_Get_file_sector(FILE *file_desc, FAT_DIR_Record *REC) {
	uint32_t cluster = (uint32_t)REC->DIR.DIR_FstClusHI;
	cluster <<= 16;
	cluster &= 0xFFFF0000;
	cluster |= (uint32_t)REC->DIR.DIR_FstClusLO;
	return(((cluster - 2) * file_desc->FAT->Param.BPB_SecPerClus) + file_desc->FAT->FirstDataSector);
	
}



static uint32_t FAT_FILE_Get_file_cluster(FILE *file_desc, FAT_DIR_Record *REC) {
	uint32_t cluster = (uint32_t)REC->DIR.DIR_FstClusHI;
	cluster <<= 16;
	cluster &= 0xFFFF0000;
	cluster |= (uint32_t)REC->DIR.DIR_FstClusLO;
	return(cluster);
	
}



static FILE_AccessType FAT_FILE_parse_fopen_mode(const char *mode) {
	FILE_AccessType AccessType = FILE_AccessType_CLOSE;
	int mode_length = strlen(mode);
	uint8_t *pointer = (uint8_t*)mode;
	for(; mode_length; mode_length--, pointer++) {
		switch(*pointer) {
			case 'r':
				AccessType |= FILE_AccessType_READ;
				break;
			case 'w':
				AccessType |= FILE_AccessType_WRITE;
				break;
			case 'a':
				AccessType |= FILE_AccessType_ADD;
				break;
			default:
				break;
		}
	}
	return(AccessType);
}



FILE *sfopen(FILE *file_desc, const char *filename, const char *mode) {
	if(file_desc && filename && mode) {
		uint8_t *path = (uint8_t*)filename;
		uint32_t path_size = strlen(filename);
		FAT_DIR_Record *REC;
		uint32_t dir_sector;
		uint8_t *result;
		uint32_t FS_counter;
		FAT_Desc *FAT;
		FILE_AccessType AccessType = FAT_FILE_parse_fopen_mode(mode);
		if(!AccessType) {
			return(0);
		}
		
		// here we check first byte of filename
		// if there is "/" character then we know, that
		// filename starts from the partition name
		if(*filename == '/') {
			path++;
			// at first we have to find which partition is it
			for(FS_counter = 0; FS_counter < bNumFATDevices; FS_counter++) {
				FAT = &FAT_Devices[FS_counter];
				result = FAT_FILE_Path_compare(path, (FAT->MSD->string_descriptor + 1));
				if(result) {
					path = result;
					FAT_FILE_Desc_init_by_root_dir(file_desc, FAT);
					break;
				}
			}
			if(FS_counter == bNumFATDevices) {
				return(0);
			}
		}
		dir_sector = file_desc->FileStartedSector;

		// here we parse path and search us file
		while(((uint8_t*)filename + path_size) > path) {
			REC = FAT_get_file_record_from_dir(file_desc, dir_sector, path, &result);
			if(REC) {
				path = result;
				if(!REC->DIR.DIR_FstClusHI && !REC->DIR.DIR_FstClusLO) {
					file_desc->ParentDirectoryStartedSector = 0;
					dir_sector = file_desc->FAT->Root_first_sector;
					FAT_FILE_Desc_init_by_root_dir(file_desc, file_desc->FAT);
					continue;
				}
				file_desc->ParentDirectoryStartedSector = dir_sector;
				dir_sector = FAT_FILE_Get_file_sector(file_desc, REC);
				FAT_FILE_Desc_init_by_dir(
						file_desc,
						REC,
						dir_sector,
						FAT_FILE_Get_file_cluster(file_desc, REC));
			}
			else {
				return(0);
			}
		}

		file_desc->AccessType = AccessType;

		if(MSMD_AccessResult_Success == MSMD_ReadBlock(file_desc->FAT->MSD, &FAT_User, (uint8_t*)file_desc->Buffer.uint8, file_desc->CurrentSector * file_desc->FAT->Param.BPB_BytsPerSec, 0, 1)) {
			file_desc->BufferPosition = 0;
			return(file_desc);
		}
		else {
			fclose(file_desc);
		}
	}
	return(0);
}




uint32_t fread(void *ptr, uint32_t size, uint32_t nobj, FILE *stream) {
	uint32_t total_size = size * nobj;
	uint32_t processed_size = total_size;

	if(!(stream->Attribute & FILE_EOF)) {
		if(total_size > (stream->FileSize - stream->FilePosition)) {
			total_size = stream->FileSize - stream->FilePosition;
			processed_size = total_size;
		}
		if(processed_size > (512 - stream->BufferPosition)) {
			memcpy(ptr, &(stream->Buffer.uint8[stream->BufferPosition]), 512 - stream->BufferPosition);
			ptr = (uint8_t*)ptr + (512 - stream->BufferPosition);
			processed_size -= 512 - stream->BufferPosition;
			if(!FAT_File_read_next_sector(stream)) {
				stream->Attribute |= FILE_EOF;
				stream->FilePosition += total_size - processed_size;
				return(total_size - processed_size);
			}
			while(processed_size >= 512) {
				memcpy(ptr, &(stream->Buffer.uint8[0]), 512);
				ptr = (uint8_t*)ptr + 512;
				processed_size -= 512;
				if(!FAT_File_read_next_sector(stream)) {
					stream->Attribute |= FILE_EOF;
					stream->FilePosition += total_size - processed_size;
					return(total_size - processed_size);
				}
			}
			
		}
		memcpy(ptr, &(stream->Buffer.uint8[stream->BufferPosition]), processed_size);
		stream->BufferPosition += processed_size;
		stream->FilePosition += total_size;
		return(total_size);
	}
	return(0);
}



int fclose (FILE *stream ) {
	FAT_DIR_Record *REC;
	uint32_t dir_sector;
	uint8_t *result;
	if(stream->AccessType && stream->ParentDirectoryStartedSector) {
		// here we get back to directory inside of us file is
		// at first we change dir_sector - it have to indicate to us direcory
		dir_sector = stream->ParentDirectoryStartedSector;
		// now we get REC structure of parent directory
		// to fill ParentDirectoryStartedSector of "stream" structure
		REC = FAT_get_file_record_from_dir(stream, dir_sector, (uint8_t*)"..", &result);
		stream->ParentDirectoryStartedSector = FAT_FILE_Get_file_sector(stream, REC);
		// here we get informations about current directory
		REC = FAT_get_file_record_from_dir(stream, dir_sector, (uint8_t*)".", &result);
		FAT_FILE_Desc_init_by_dir(
				stream,
				REC,
				dir_sector,
				FAT_FILE_Get_file_cluster(stream, REC));
		if(MSMD_AccessResult_Success == MSMD_ReadBlock(stream->FAT->MSD, &FAT_User, (uint8_t*)stream->Buffer.uint8, stream->CurrentSector * stream->FAT->Param.BPB_BytsPerSec, 0, 1)) {
			stream->BufferPosition = 0;
			return(0);
		}
	}
	return(-1);
}




void FAT_Init(void) {
	FAT_User.MSMD_ConnectEventHandler = FAT_ConnectEvent;

	FAT_CACHE_SEMAPHORE_INIT();
	FAT_LFN_SEMAPHORE_INIT();

	MSMU_Install(&FAT_User);
}


static FAT32_BPB FAT;
static void FAT_ConnectEvent (MSMD_Desc *MSD) {
	uint32_t counter;
	FAT_Desc *FATD;
	FAT32_BPB_Special *FAT_Spec;
	uint32_t RootDir_sectors_count;
	uint32_t data_sectors_count;
//	int i,j;
	FAT_CACHE_SEMAPHORE_GET();
	
	FAT_Cache_stored_sector = 0xFFFFFFFF;
	FAT_Cache_owner = FAT_MAX_NUM_DEVICES + 1;

	if(MSMD_AccessResult_Success == MSMD_ReadBlock(MSD, &FAT_User, (uint8_t*)(FAT_Cache.uint8), 0, 0, 1)) {
/*
		for(i=0;i<512;) {
			print_string((uint8_t*)"\n\r");
			for(j=0; j < 32; j++,i++) {
				print_string((uint8_t*)" ");
				print_data(FAT_Cache.uint8[i], 2, 0, 0);
			}
		}
		print_string((uint8_t*)"\n\r");//*/
		
		FAT.BPB_BytsPerSec = ((uint16_t)FAT_Cache.uint8[11]) & 0xFF;
		FAT.BPB_BytsPerSec |= (((uint16_t)FAT_Cache.uint8[12]) << 8) & 0xFF00;

		FAT.BPB_SecPerClus = FAT_Cache.uint8[13];

		FAT.BPB_RsvdSecCnt = FAT_Cache.uint16[14 / sizeof(uint16_t)];

		FAT.BPB_NumFATs = FAT_Cache.uint8[16];

		FAT.BPB_RootEntCnt = ((uint16_t)FAT_Cache.uint8[17]) & 0xFF;
		FAT.BPB_RootEntCnt |= (((uint16_t)FAT_Cache.uint8[18]) << 8) & 0xFF00;

		FAT.BPB_TotSec16 = ((uint16_t)FAT_Cache.uint8[19]) & 0xFF;
		FAT.BPB_TotSec16 |= (((uint16_t)FAT_Cache.uint8[20]) << 8) & 0xFF00;

		FAT.BPB_Media = FAT_Cache.uint8[21];

		FAT.BPB_FATSz16 = FAT_Cache.uint16[22 / sizeof(uint16_t)];

		FAT.BPB_SecPerTrk = FAT_Cache.uint16[24 / sizeof(uint16_t)];

		FAT.BPB_NumHeads = FAT_Cache.uint16[26 / sizeof(uint16_t)];

		FAT.BPB_HiddSec = FAT_Cache.uint32[28 / sizeof(uint32_t)];

		FAT.BPB_TotSec32 = FAT_Cache.uint32[32 / sizeof(uint32_t)];

		FATD = &FAT_Devices[bNumFATDevices];
		FATD->FAT_ID = bNumFATDevices;
		FATD->FAT_name_string_descriptor = MSD->string_descriptor;
		bNumFATDevices++;
		FATD->MSD = MSD;

		// FATSz16
		if(FAT.BPB_FATSz16) {
			FATD->FATSz = (uint32_t)(FAT.BPB_FATSz16);
		}
		else {
			FATD->FATSz = FAT_Cache.uint32[36 / sizeof(uint32_t)];
		}
		
		// TotSec16
		if(FAT.BPB_TotSec16) {
			FATD->TotSec = (uint32_t)(FAT.BPB_TotSec16);
		}
		else {
			FATD->TotSec = (uint32_t)(FAT.BPB_TotSec32);
		}

		RootDir_sectors_count = ((FAT.BPB_RootEntCnt << 5)
				+ (FAT.BPB_BytsPerSec - 1)) / FAT.BPB_BytsPerSec;
		data_sectors_count = FATD->TotSec - (FAT.BPB_RsvdSecCnt
				+ (FAT.BPB_NumFATs * FATD->FATSz) + RootDir_sectors_count);
		data_sectors_count /= FAT.BPB_SecPerClus;
		if(data_sectors_count < 4085) {
			FATD->FAT_Type = 12;
		}
		else if(data_sectors_count < 65525) {
			FATD->FAT_Type = 16;
		}
		else {
			FATD->FAT_Type = 32;
		}
		
		FATD->FirstDataSector = FAT.BPB_RsvdSecCnt + (FAT.BPB_NumFATs * FATD->FATSz) + RootDir_sectors_count;

		FATD->Param.BPB_BytsPerSec = FAT.BPB_BytsPerSec;
		FATD->Param.BPB_SecPerClus = FAT.BPB_SecPerClus;
		FATD->Param.BPB_RsvdSecCnt = FAT.BPB_RsvdSecCnt;
		FATD->Param.BPB_RootEntCnt = FAT.BPB_RootEntCnt;
		
		FAT_Spec = (FAT32_BPB_Special*)(&(FAT_Cache.uint8[36]));

		
		FATD->Param.BPB_RootClus = FAT_Spec->BPB_RootClus;
		
		FATD->Root_first_sector = (((FAT_Spec->BPB_RootClus - 2) * FAT.BPB_SecPerClus) + FATD->FirstDataSector);// * FAT.BPB_BytsPerSec;

		FAT_CACHE_SEMAPHORE_RETURN();

		for(counter = 0; counter < bNumFATUsers; counter++) {
			if(FAT_Users[counter]->FAT_ConnectEventHandler) {
				FAT_Users[counter]->FAT_ConnectEventHandler(FATD);
			}
		}
		return;
	}
	FAT_CACHE_SEMAPHORE_RETURN();
}



int FAT_UserInstall(FAT_UserDesc *FATU) {
	if(bNumFATUsers < FAT_MAX_NUM_USERS && FATU) {
		FAT_Users[bNumFATUsers] = FATU;
		bNumFATUsers++;
		return(0);
	}
	return(-1);
}
