/*
================================================================================
 File: fat32.h
--------------------------------------------------------------------------------
 Author:        Piotr Wojtowicz
 Purpose:       FAT16/32 read-only library MSM compatible
 Compiler:      ANSI C
 Hardware:      
 Doc:
--------------------------------------------------------------------------------
 Includes:      
--------------------------------------------------------------------------------
 Comments:                      UWAGA!!!
                Oprogramowanie to udostępnione jest do celów poglądowych,
                nie jest przetestowane pod kątem bezpieczeństwa ani spełnienia
                jakichkolwiek norm wymaganych aby urządzenie używające go
                mogło być dopuszczone do użytku, dlatego używasz go
                na własną odpowiedzialność a jego autor nie ponosi żadnego
                typu odpowiedzialności za jakiekolwiek skutki jego użycia.
                Jeżeli nie zgadzasz się z powyższym warunkiem proszę
                o nieużywanie tego oprogramowania.
================================================================================
*/

#ifndef FAT32_H_
#define FAT32_H_
#include <stdint.h>
#include "msm.h"



#define FAT_MAX_NUM_DEVICES					MSM_MAX_NUM_DEVICES

#define FAT_MAX_NUM_USERS					1

#define FILE_ALLOC_TABLE_BUFFER_SIZE		16

#define OS_SUPPORT							0


// UWAGA!!! nie mapowac tych struktur na zadne tablice
// odczytane bezposrednio z pamieci gdyz moze to powodowac
// w zaleznosci od platformy przesuniecia i zla interpretacje
// odebranych z pamieci danych
// Jest to spowodowane nieprzemyslanym ustawieniem pol
// wewnatrz struktury BPB w standardzie FAT i wyrownywaniem
// przez kompilatory elementow struktur do adresow
// slow 16- i 32-bitowych.
typedef struct {
	uint8_t BS_jmpBoot[3];
	uint8_t BS_OEM_Name[8];
	uint16_t BPB_BytsPerSec;
	uint8_t BPB_SecPerClus;
	uint16_t BPB_RsvdSecCnt;
	uint8_t BPB_NumFATs;
	uint16_t BPB_RootEntCnt;
	uint16_t BPB_TotSec16;
	uint8_t BPB_Media;
	uint16_t BPB_FATSz16;
	uint16_t BPB_SecPerTrk;
	uint16_t BPB_NumHeads;
	uint32_t BPB_HiddSec;
	uint32_t BPB_TotSec32;
}FAT32_BPB;

typedef struct {
	uint8_t BS_DrvNum;
	uint8_t BS_Reserved1;
	uint8_t BS_BootSig;
	uint32_t BS_VolID;
	uint8_t BS_VolLab[11];
	uint8_t BS_FilSysType[8];
}FAT16_BPB_Special;

typedef struct {
	uint32_t BPB_FATSz32;
	uint16_t BPB_ExtFlags;
	uint16_t BPB_FSVer;
	uint32_t BPB_RootClus;
	uint16_t BPB_FS_Info;
	uint16_t BPB_BkBootSec;
	uint8_t BPB_Reserved[12];
	FAT16_BPB_Special FAT16_BPB;
}FAT32_BPB_Special;


typedef struct {
	uint8_t BPB_SecPerClus;
	uint16_t BPB_BytsPerSec;
	uint16_t BPB_RsvdSecCnt;
	uint16_t BPB_RootEntCnt;
	uint32_t BPB_RootClus;
}BPB_FAT;


typedef struct {
	uint32_t FAT_ID;
	MSMD_Desc *MSD;
	uint32_t TotSec;
	uint32_t FATSz;
	uint32_t FirstDataSector;
	uint32_t Root_first_sector;
	uint8_t *FAT_name_string_descriptor;
	uint8_t FAT_Type;
	BPB_FAT Param;
} FAT_Desc;



typedef struct {
	uint8_t DIR_Name[11];
	uint8_t DIR_Attr;
	uint8_t DIR_NTRes;
	uint8_t DIR_CrtTimeTenth;
	uint16_t DIR_CrtTime;
	uint16_t DIR_CrtDate;
	uint16_t DIR_LstAccDate;
	uint16_t DIR_FstClusHI;
	uint16_t DIR_WrtTime;
	uint16_t DIR_WrtDate;
	uint16_t DIR_FstClusLO;
	uint32_t DIR_FileSize;
}FAT_DirectoryRecord;

#define ATTR_READ_ONLY		(0x1 << 0)
#define ATTR_HIDDEN			(0x1 << 1)
#define ATTR_SYSTEM			(0x1 << 2)
#define ATTR_VOLUME_ID		(0x1 << 3)
#define ATTR_DIRECTORY		(0x1 << 4)
#define ATTR_ARCHIVE		(0x1 << 5)
#define ATTR_LONG_NAME		(ATTR_READ_ONLY | ATTR_HIDDEN | ATTR_SYSTEM | ATTR_VOLUME_ID)


typedef struct {
	uint8_t LFN_OrdinalField;
	uint8_t LFN_NameChars1[10];
	uint8_t DIR_Attr;
	uint8_t LFN_Type;
	uint8_t LFN_Checksum;
	uint8_t LFN_NameChars2[12];
	uint16_t DIR_ClusLO;
	uint8_t LFN_NameChars3[4];
}FAT_LFN_Record;



typedef union {
	FAT_DirectoryRecord DIR;
	FAT_LFN_Record LFN;
}FAT_DIR_Record;



typedef enum {
	FILE_AccessType_CLOSE   = 0,
	FILE_AccessType_READ    = 1 << 0,
	FILE_AccessType_WRITE   = 1 << 1,
	FILE_AccessType_ADD     = 1 << 2
}FILE_AccessType;

typedef struct {
	FAT_Desc *FAT;
	uint32_t ParentDirectoryStartedSector;
	uint32_t FileStartedSector;
	uint32_t CurrentSector;
	uint32_t AllocTableBuffer[FILE_ALLOC_TABLE_BUFFER_SIZE];
	uint16_t AllocTableBufferPosition;
	uint16_t AllocTableBufferStoredDataSize;
	uint32_t FileSize;
	uint32_t FilePosition;
	uint16_t BufferPosition;
	union {
		uint8_t uint8[512 / sizeof(uint8_t)];
		uint16_t uint16[512 / sizeof(uint16_t)];
		uint32_t uint32[512 / sizeof(uint32_t)];
	}Buffer;
//	uint8_t Buffer[512];
	FILE_AccessType AccessType;
	uint8_t Attribute;
}FILE;

#define FILE_EOF				(1 << 0)
#define FILE_TYPE_DIRECTORY		(1 << 1)





typedef struct {
	// wskaznik do funkcji zainstalowanego uzytkownika
	// partycji wywolywana w momencie wykrycia w systemie
	// nowego nosnika pamieci masowej
	void (*FAT_ConnectEventHandler)(FAT_Desc *FATD);

	// wskaznik do funkcji zainstalowanego uzytkownika
	// pamieci wywolywana w momencie odlaczenia z systemu
	// danego nosnika pamieci
	void (*FAT_DisconnectEventHandler)(FAT_Desc *FATD);
}FAT_UserDesc;





FILE *sfopen(FILE *file_desc, const char *filename, const char *mode);

uint32_t fread(void *ptr, uint32_t size, uint32_t nobj, FILE *stream);

int fclose (FILE *stream );

int FAT_UserInstall(FAT_UserDesc *FATU);

void FAT_Init(void);



#endif /*FAT32_H_*/
