/*
================================================================================
 File: msm.c
--------------------------------------------------------------------------------
 Author:        Piotr Wojtowicz
 Purpose:       Mass Storage Manager library version 1.0
 Compiler:      ANSI C
 Hardware:      
 Doc:
--------------------------------------------------------------------------------
 Includes:      
--------------------------------------------------------------------------------
 Comments:                      UWAGA!!!
                Oprogramowanie to udostępnione jest do celów poglądowych,
                nie jest przetestowane pod kątem bezpieczeństwa ani spełnienia
                jakichkolwiek norm wymaganych aby urządzenie używające go
                mogło być dopuszczone do użytku, dlatego używasz go
                na własną odpowiedzialność a jego autor nie ponosi żadnego
                typu odpowiedzialności za jakiekolwiek skutki jego użycia.
                Jeżeli nie zgadzasz się z powyższym warunkiem proszę
                o nieużywanie tego oprogramowania.
================================================================================
*/

#include "msm.h"
#include <string.h>


static MSMD_Desc *Devices[MSM_MAX_NUM_DEVICES];
static MSMD_AccessParameters *Parameters[MSM_MAX_NUM_DEVICES];
static uint32_t bNumInstalledDevices = 0;

static MSMU_Desc *Users[MSM_MAX_NUM_USERS];
static uint32_t bNumInstalledUsers = 0;



int MSMD_Install(MSMD_Desc *MSD, MSMD_AccessParameters *MSDAP) {
	if((bNumInstalledDevices < MSM_MAX_NUM_DEVICES) && MSD && MSDAP) {
		uint32_t counter;
		for(counter = 0; counter < bNumInstalledDevices; counter++) {
			if(Devices[counter] == MSD) {
				return(-1);
			}
		}
		MSD->ID = bNumInstalledDevices;
#if((MSM_READ_SUPPORT == 1) || (MSM_VERIFY_SUPPORT == 1))
		MSDAP->READ_VERIFY_owner = 0;
#endif
#if((MSM_WRITE_SUPPORT == 1) || (MSM_ERASE_SUPPORT == 1))
		MSDAP->WRITE_ERASE_owner = 0;
#endif
		MSDAP->DeviceState = MSMD_AccessResult_Idle;
		Devices[bNumInstalledDevices] = MSD;
		Parameters[bNumInstalledDevices] = MSDAP;
		bNumInstalledDevices++;

		for(counter = 0; counter < bNumInstalledUsers; counter++) {
			if(Users[counter]->MSMD_ConnectEventHandler) {
				Users[counter]->MSMD_ConnectEventHandler(MSD);
			}
		}
		return(0);
	}
	return(-1);
}



#if(MSM_MULTI_SESSION_SUPPORT == 1)
int MSMD_Uninstall(MSMD_Desc *MSD) {
	if(MSD) {
		uint32_t counter;
		for(counter = 0; counter < bNumInstalledDevices; counter++) {
			if(Devices[counter] == MSD) {
				break;
			}
		}
		if(counter < bNumInstalledDevices) {
			bNumInstalledDevices--;
			while(counter < bNumInstalledDevices) {
				Devices[counter] = Devices[counter + 1];
				Parameters[counter] = Parameters[counter + 1];
			}

			for(counter = 0; counter < bNumInstalledUsers; counter++) {
				if(Users[counter]->MSMD_DisconnectEventHandler) {
					Users[counter]->MSMD_DisconnectEventHandler(MSD);
				}
			}
		}
		return(0);
	}
	return(-1);
}
#endif



#if(MSM_GET_DEVICES_SUPPORT == 1)
uint32_t MSM_GetNumInstalledDevices(void) {
	return(bNumInstalledDevices);
}



MSMD_Desc **MSM_GetDevicesTable(void) {
	return(Devices);
}



MSMD_Desc * MSM_GetDeviceByName(const char *name) {
	uint32_t counter;
	uint8_t *devname;
	for(counter = 0; counter < bNumInstalledDevices; counter++) {
		devname = (uint8_t*)Devices[counter]->string_descriptor;
		if(*name != '/') {
			devname++;
		}
		if(!strcmp(name, (const char*)devname)) {
			return(Devices[counter]);
		}
	}
	return(0);
}



MSMD_Desc * MSM_GetDeviceByID(uint32_t id) {
	if(id < bNumInstalledDevices) {
		return(Devices[id]);
	}
	return(0);
}
#endif



#if(MSM_USERS_SUPPORT == 1)
int MSMU_Install(MSMU_Desc *MSU) {
	if((bNumInstalledUsers < MSM_MAX_NUM_USERS) && MSU) {
		uint32_t counter;
		for(counter = 0; counter < bNumInstalledUsers; counter++) {
			if(Users[counter] == MSU) {
				return(-1);
			}
		}
		Users[bNumInstalledUsers] = MSU;
		bNumInstalledUsers++;

		if(MSU->MSMD_ConnectEventHandler) {
			for(counter = 0; counter < bNumInstalledDevices; counter++) {
				MSU->MSMD_ConnectEventHandler(Devices[counter]);
			}
		}
		return(0);
	}
	return(-1);
}



#if(MSM_MULTI_SESSION_SUPPORT == 1)
int MSMU_Uninstall(MSMU_Desc *MSU) {
	if(MSU) {
		uint32_t counter;
		for(counter = 0; counter < bNumInstalledUsers; counter++) {
			if(Users[counter] == MSU) {
				break;
			}
		}
		bNumInstalledUsers--;
		while(counter < bNumInstalledUsers) {
			Users[counter] = Users[counter + 1];
		}

		if(MSU->MSMD_DisconnectEventHandler) {
			for(counter = 0; counter < bNumInstalledDevices; counter++) {
				MSU->MSMD_DisconnectEventHandler(Devices[counter]);
			}
		}
		return(0);
	}
	return(-1);
}
#endif
#endif



#if(MSM_READ_SUPPORT == 1)
MSMD_AccessResult MSMD_ReadBlock(MSMD_Desc *MSD, MSMU_Desc *MSU, uint8_t *data, uint32_t addr_L, uint32_t addr_H, uint32_t count) {
	if(MSD && count) {
		if((MSD == (Devices[MSD->ID])) && !(addr_L % MSD->read_sector_size)) {
			MSMD_AccessParameters *AP = Parameters[MSD->ID];
#if(MSM_USERS_SUPPORT == 1)
			if(!(AP->READ_VERIFY_owner) || ((MSMU_Desc *)(AP->READ_VERIFY_owner) == MSU)) {
#endif
				if(AP->DeviceState == MSMD_AccessResult_Idle) {
					return(AP->ReadBlockHandler(MSD, data, addr_L, addr_H, count));
				}
				else {
					return(MSMD_AccessResult_Busy);
				}
#if(MSM_USERS_SUPPORT == 1)
			}
#endif
		}
	}
	return(MSMD_AccessResult_AccessForbidden);
}
#endif



#if(MSM_WRITE_SUPPORT == 1)
MSMD_AccessResult MSMD_WriteBlock(MSMD_Desc *MSD, MSMU_Desc *MSU, uint8_t *data, uint32_t addr_L, uint32_t addr_H, uint32_t count) {
	if(MSD && count) {
		if((MSD == (Devices[MSD->ID])) && !(addr_L % MSD->write_sector_size)) {
			MSMD_AccessParameters *AP = Parameters[MSD->ID];
#if(MSM_USERS_SUPPORT == 1)
			if(!(AP->WRITE_ERASE_owner) || ((MSMU_Desc *)(AP->WRITE_ERASE_owner) == MSU)) {
#endif
				if(AP->DeviceState == MSMD_AccessResult_Idle) {
					return(AP->WriteBlockHandler(MSD, data, addr_L, addr_H, count));
				}
				else {
					return(MSMD_AccessResult_Busy);
				}
#if(MSM_USERS_SUPPORT == 1)
			}
#endif
		}
	}
	return(MSMD_AccessResult_AccessForbidden);
}
#endif



#if(MSM_VERIFY_SUPPORT == 1)
MSMD_AccessResult MSMD_VerifyBlock(MSMD_Desc *MSD, MSMU_Desc *MSU, uint32_t addr_L, uint32_t addr_H, uint32_t count) {
	if(MSD && count) {
		if((MSD == (Devices[MSD->ID])) && !(addr_L % MSD->verify_sector_size)) {
			MSMD_AccessParameters *AP = Parameters[MSD->ID];
#if(MSM_USERS_SUPPORT == 1)
			if(!(AP->READ_VERIFY_owner) || ((MSMU_Desc *)(AP->READ_VERIFY_owner) == MSU)) {
#endif
				if(AP->DeviceState == MSMD_AccessResult_Idle) {
					return(AP->VerifyBlockHandler(MSD, addr_L, addr_H, count));
				}
				else {
					return(MSMD_AccessResult_Busy);
				}
#if(MSM_USERS_SUPPORT == 1)
			}
#endif
		}
	}
	return(MSMD_AccessResult_AccessForbidden);
}
#endif



#if(MSM_ERASE_SUPPORT == 1)
MSMD_AccessResult MSMD_EraseBlock(MSMD_Desc *MSD, MSMU_Desc *MSU, uint32_t addr_L, uint32_t addr_H, uint32_t count) {
	if(MSD && count) {
		if((MSD == (Devices[MSD->ID])) && !(addr_L % MSD->erase_sector_size)) {
			MSMD_AccessParameters *AP = Parameters[MSD->ID];
#if(MSM_USERS_SUPPORT == 1)
			if(!(AP->WRITE_ERASE_owner) || ((MSMU_Desc *)(AP->WRITE_ERASE_owner) == MSU)) {
#endif
				if(AP->DeviceState == MSMD_AccessResult_Idle) {
					return(AP->EraseBlockHandler(MSD, addr_L, addr_H, count));
				}
				else {
					return(MSMD_AccessResult_Busy);
				}
#if(MSM_USERS_SUPPORT == 1)
			}
#endif
		}
	}
	return(MSMD_AccessResult_AccessForbidden);
}
#endif
