/*
================================================================================
 File: msm.h
--------------------------------------------------------------------------------
 Author:        Piotr Wojtowicz
 Purpose:       Mass Storage Manager library version 1.0
 Compiler:      ANSI C
 Hardware:      
 Doc:
--------------------------------------------------------------------------------
 Includes:      
--------------------------------------------------------------------------------
 Comments:                      UWAGA!!!
                Oprogramowanie to udostępnione jest do celów poglądowych,
                nie jest przetestowane pod kątem bezpieczeństwa ani spełnienia
                jakichkolwiek norm wymaganych aby urządzenie używające go
                mogło być dopuszczone do użytku, dlatego używasz go
                na własną odpowiedzialność a jego autor nie ponosi żadnego
                typu odpowiedzialności za jakiekolwiek skutki jego użycia.
                Jeżeli nie zgadzasz się z powyższym warunkiem proszę
                o nieużywanie tego oprogramowania.
================================================================================
*/

#ifndef MSM_H_
#define MSM_H_
#include <stdint.h>



typedef enum {
	MSMD_AccessResult_Idle = 0,
	MSMD_AccessResult_Success,
	MSMD_AccessResult_Busy,
	MSMD_AccessResult_DataError,
	MSMD_AccessResult_AccessForbidden
}MSMD_AccessResult;


#define MSM_MAX_NUM_DEVICES         1
#define MSM_MAX_NUM_USERS           1

#define MSM_READ_SUPPORT            1

#define MSM_WRITE_SUPPORT           0

#define MSM_VERIFY_SUPPORT          0

#define MSM_ERASE_SUPPORT           0

#define MSM_MULTI_SESSION_SUPPORT   0

#define MSM_GET_DEVICES_SUPPORT     1

#define MSM_USERS_SUPPORT           1


typedef struct {
	// numer identyfikcyjny pamieci masowej
	uint32_t ID;

	// wskaznik do stringa opisujacy dana pamiec
	uint8_t *string_descriptor;

	// mlodsza polowka rozmiaru pamieci
	uint32_t area_size_L;
	// starsza polowka rozmiaru pamieci
	uint32_t area_size_H;

#if(MSM_READ_SUPPORT == 1)
	// rozmiar sektora pamieci do odczytu
	uint32_t read_sector_size;
#endif

#if(MSM_WRITE_SUPPORT == 1)
	// rozmiar sektora pamieci do zapisu
	uint32_t write_sector_size;
#endif

#if(MSM_VERIFY_SUPPORT == 1)
	// rozmiar sektora pamieci do weryfikacji
	uint32_t verify_sector_size;
#endif

#if(MSM_ERASE_SUPPORT == 1)
	// rozmiar sektora pamieci do kasowania
	uint32_t erase_sector_size;
#endif
}MSMD_Desc;


// ta struktura jest tylko do uzytku portu i biblioteki mass_storage_manager
// uzytkownik nie ma do niej dostepu
typedef struct {
#if(MSM_READ_SUPPORT == 1)
	// wskaznik do funkcji odczytu z nosnika
	MSMD_AccessResult (*ReadBlockHandler)(MSMD_Desc *MSD, uint8_t *data, uint32_t addr_L, uint32_t addr_H, uint32_t count);
#endif

#if(MSM_WRITE_SUPPORT == 1)
	// wskaznik do funckji zapisu do nosnika
	MSMD_AccessResult (*WriteBlockHandler)(MSMD_Desc *MSD, uint8_t *data, uint32_t addr_L, uint32_t addr_H, uint32_t count);
#endif

#if(MSM_VERIFY_SUPPORT == 1)
	// wskaznik do funkcji weryfikacyjnej dla danego nosnika
	MSMD_AccessResult (*VerifyBlockHandler)(MSMD_Desc *MSD, uint32_t addr_L, uint32_t addr_H, uint32_t count);
#endif

#if(MSM_ERASE_SUPPORT == 1)
	// wskaznik do funkcji kasujacej dla danego nosnika
	MSMD_AccessResult (*EraseBlockHandler)(MSMD_Desc *MSD, uint32_t addr_L, uint32_t addr_H, uint32_t count);
#endif

#if(MSM_USERS_SUPPORT == 1)
#if((MSM_READ_SUPPORT == 1) || (MSM_VERIFY_SUPPORT == 1))
	// wskaznik do wlasciciela cech READ i VERIFY
	// jezeli ten wskaznik jest wyzerowany wowczas
	// kazdy moze wykonywac dane operacje na nosniku
	// w przypadku niezerowej wartosci tego wskaznika
	// czytac z nosnika i weryfikowac zawartosc moze jedynie
	// obecny wlasciciej tej cechy
	void *READ_VERIFY_owner;
#endif

#if((MSM_WRITE_SUPPORT == 1) || (MSM_ERASE_SUPPORT == 1))
	// wskaznik do wlasciciela cech WRITE i ERASE
	// ten wskaznik musi byc ustawiony - powienien
	// istniec wlasciciel tych cech nosnika
	// w przeciwnym przypadku istnieje zagrozenie
	// nadpisywania danych przez roznych uzytkownikow
	// i w konsekwencji wystapienie bledow na nosniku
	void *WRITE_ERASE_owner;
#endif
#endif

	// informuje, czy sa aktualnie prowadzone jakies
	// operacje na nosniku - mozliwe opcje: Idle i Busy
	// UWAGA!!!
	// tym polem "opiekuje się" biblioteka portu
	// dzieki temu jezeli jest to mozliwe biblioteka portu
	// moze prowadzic kilka transmisji rownoczesnie
	MSMD_AccessResult DeviceState;
}MSMD_AccessParameters;


typedef struct {
	// wskaznik do funkcji zainstalowanego uzytkownika
	// pamieci wywolywana w momencie wykrycia w systemie
	// nowego nosnika pamieci masowej
	void (*MSMD_ConnectEventHandler)(MSMD_Desc *MSD);

	// wskaznik do funkcji zainstalowanego uzytkownika
	// pamieci wywolywana w momencie odlaczenia z systemu
	// danego nosnika pamieci
	void (*MSMD_DisconnectEventHandler)(MSMD_Desc *MSD);

#if((MSM_READ_SUPPORT == 1) || (MSM_VERIFY_SUPPORT == 1))
	// wskaznik do fukcji uwolnienia cech READ i VERIFY nosnika
	// funkcja ta wywolywana jest gdy dane cechy pamieci
	// sa zajete przez jakiegos uzytkownika a inny uzytkownik
	// potrzebuje skorzystac z danych cech nosnika.
	// wywolujac ta funkcje "prosimy" obecnego wlasciciela
	// cechy nosnika aby zakonczyl swoje operacje na nosniku
	// i zwolnil cechy
	void (*MSMD_READ_VERIFY_ReleaseHandler)(MSMD_Desc *MSD);
#endif

#if((MSM_WRITE_SUPPORT == 1) || (MSM_ERASE_SUPPORT == 1))
	// wskaznik do fukcji uwolnienia cech WRITE i ERASE nosnika
	// znaczenie takie samo jak MSD_READ_VERIFY_release
	// z ta roznica, ze dotyczy cech WRITE i ERASE
	void (*MSMD_WRITE_ERASE_ReleaseHandler)(MSMD_Desc *MSD);
#endif
}MSMU_Desc;




// mozliwy jedynie odczyt blokow
// adresacja odnosi sie do bajtow a nie sektorow - jednostka adresowa jest bajt
// przy czym adres moze byc jedynie calkowita wielokrotnoscia pojemnosci sektora
// count jest natomiast liczba sektorow
// wskaznik do danych powinien byc adresem podzielnym przez 4 ale nie jest to wymog
#if(MSM_READ_SUPPORT == 1)
MSMD_AccessResult MSMD_ReadBlock(MSMD_Desc *MSD, MSMU_Desc *MSU, uint8_t *data, uint32_t addr_L, uint32_t addr_H, uint32_t count);
#endif

#if(MSM_WRITE_SUPPORT == 1)
MSMD_AccessResult MSMD_WriteBlock(MSMD_Desc *MSD, MSMU_Desc *MSU, uint8_t *data, uint32_t addr_L, uint32_t addr_H, uint32_t count);
#endif

#if(MSM_VERIFY_SUPPORT == 1)
MSMD_AccessResult MSMD_VerifyBlock(MSMD_Desc *MSD, MSMU_Desc *MSU, uint32_t addr_L, uint32_t addr_H, uint32_t count);
#endif

#if(MSM_ERASE_SUPPORT == 1)
MSMD_AccessResult MSMD_EraseBlock(MSMD_Desc *MSD, MSMU_Desc *MSU, uint32_t addr_L, uint32_t addr_H, uint32_t count);
#endif

int MSMD_Install(MSMD_Desc *MSD, MSMD_AccessParameters *MSDAP);
#if(MSM_MULTI_SESSION_SUPPORT == 1)
int MSMD_Uninstall(MSMD_Desc *MSD);
#endif

#if(MSM_GET_DEVICES_SUPPORT == 1)
uint32_t MSM_GetNumInstalledDevices(void);
MSMD_Desc ** MSM_GetDevicesTable(void);
MSMD_Desc * MSM_GetDeviceByName(const char *name);
MSMD_Desc * MSM_GetDeviceByID(uint32_t id);
#endif

#if(MSM_USERS_SUPPORT == 1)
int MSMU_Install(MSMU_Desc *MSU);
#if(MSM_MULTI_SESSION_SUPPORT == 1)
int MSMU_Uninstall(MSMU_Desc *MSU);
#endif
#endif




#endif /*MSM_H_*/
