#include <stdint.h>

#include <stdint.h>
#include <string.h>

#include "stm32f10x_map.h"
#include "stm32f10x_gpio.h"
#include "stm32f10x_rcc.h"
#include "stm32f10x_spi.h"
#include "stm32f10x_usart.h"
#include "stm32f10x_flash.h"


#if defined(OLIMEX_P103) && (OLIMEX_P103 != 0)

#define LED_PIN		GPIO_Pin_12
#define LED_PORT	GPIOC
#define BTN_PIN		GPIO_Pin_0
#define BTN_PORT	GPIOA

void LedOn(void);
void LedOff(void);
uint32_t BtnState(void);
void BtnWaitPress(void);

#endif

void MCUBusyDelay(uint32_t delay);
void CommHwSendByte(uint8_t c);


/* ------------ RCC registers bit address in the alias region ----------- */
#define RCC_OFFSET                (RCC_BASE - PERIPH_BASE)

/* --- CR Register ---*/
/* Alias word address of HSION bit */
#define CR_OFFSET                 (RCC_OFFSET + 0x00)
#define HSION_BitNumber           0x00
#define CR_HSION_BB               (PERIPH_BB_BASE + (CR_OFFSET * 32) + (HSION_BitNumber * 4))

/* Alias word address of PLLON bit */
#define PLLON_BitNumber           0x18
#define CR_PLLON_BB               (PERIPH_BB_BASE + (CR_OFFSET * 32) + (PLLON_BitNumber * 4))

/* Alias word address of CSSON bit */
#define CSSON_BitNumber           0x13
#define CR_CSSON_BB               (PERIPH_BB_BASE + (CR_OFFSET * 32) + (CSSON_BitNumber * 4))

/* --- CFGR Register ---*/
/* Alias word address of USBPRE bit */
#define CFGR_OFFSET               (RCC_OFFSET + 0x04)
#define USBPRE_BitNumber          0x16
#define CFGR_USBPRE_BB            (PERIPH_BB_BASE + (CFGR_OFFSET * 32) + (USBPRE_BitNumber * 4))

/* --- BDCR Register ---*/
/* Alias word address of RTCEN bit */
#define BDCR_OFFSET               (RCC_OFFSET + 0x20)
#define RTCEN_BitNumber           0x0F
#define BDCR_RTCEN_BB             (PERIPH_BB_BASE + (BDCR_OFFSET * 32) + (RTCEN_BitNumber * 4))

/* Alias word address of BDRST bit */
#define BDRST_BitNumber           0x10
#define BDCR_BDRST_BB             (PERIPH_BB_BASE + (BDCR_OFFSET * 32) + (BDRST_BitNumber * 4))

/* --- CSR Register ---*/
/* Alias word address of LSION bit */
#define CSR_OFFSET                (RCC_OFFSET + 0x24)
#define LSION_BitNumber           0x00
#define CSR_LSION_BB              (PERIPH_BB_BASE + (CSR_OFFSET * 32) + (LSION_BitNumber * 4))

/* ---------------------- RCC registers bit mask ------------------------ */
/* CR register bit mask */
#define CR_HSEBYP_Reset           ((u32)0xFFFBFFFF)
#define CR_HSEBYP_Set             ((u32)0x00040000)
#define CR_HSEON_Reset            ((u32)0xFFFEFFFF)
#define CR_HSEON_Set              ((u32)0x00010000)
#define CR_HSITRIM_Mask           ((u32)0xFFFFFF07)

/* CFGR register bit mask */
#define CFGR_PLL_Mask             ((u32)0xFFC0FFFF)
#define CFGR_PLLMull_Mask         ((u32)0x003C0000)
#define CFGR_PLLSRC_Mask          ((u32)0x00010000)
#define CFGR_PLLXTPRE_Mask        ((u32)0x00020000)
#define CFGR_SWS_Mask             ((u32)0x0000000C)
#define CFGR_SW_Mask              ((u32)0xFFFFFFFC)
#define CFGR_HPRE_Reset_Mask      ((u32)0xFFFFFF0F)
#define CFGR_HPRE_Set_Mask        ((u32)0x000000F0)
#define CFGR_PPRE1_Reset_Mask     ((u32)0xFFFFF8FF)
#define CFGR_PPRE1_Set_Mask       ((u32)0x00000700)
#define CFGR_PPRE2_Reset_Mask     ((u32)0xFFFFC7FF)
#define CFGR_PPRE2_Set_Mask       ((u32)0x00003800)
#define CFGR_ADCPRE_Reset_Mask    ((u32)0xFFFF3FFF)
#define CFGR_ADCPRE_Set_Mask      ((u32)0x0000C000)

/* CSR register bit mask */
#define CSR_RMVF_Set              ((u32)0x01000000)

/* RCC Flag Mask */
#define FLAG_Mask                 ((u8)0x1F)

/* Typical Value of the HSI in Hz */
#define HSI_Value                 ((u32)8000000)

/* BDCR register base address */
#define BDCR_BASE                 (PERIPH_BASE + BDCR_OFFSET)

/* Time out for HSE start up */
#define HSEStartUp_TimeOut        ((u8)0xFF)


#define AIRCR_VECTKEY_MASK    ((u32)0x05FA0000)


/* Flash Access Control Register bits */
#define ACR_PRFTBS_Mask          ((u32)0x00000020) 

/* Flash Control Register bits */
#define CR_PG_Set                ((u32)0x00000001)
#define CR_PG_Reset              ((u32)0x00001FFE) 

#define CR_PER_Set               ((u32)0x00000002)
#define CR_PER_Reset             ((u32)0x00001FFD)

#define CR_MER_Set               ((u32)0x00000004)
#define CR_MER_Reset             ((u32)0x00001FFB)

#define CR_OPTPG_Set             ((u32)0x00000010)
#define CR_OPTPG_Reset           ((u32)0x00001FEF)

#define CR_OPTER_Set             ((u32)0x00000020)
#define CR_OPTER_Reset           ((u32)0x00001FDF)

#define CR_STRT_Set              ((u32)0x00000040)
							 
#define CR_LOCK_Set              ((u32)0x00000080)

/* FLASH Mask */
#define RDPRT_Mask               ((u32)0x00000002)
#define WRP0_Mask                ((u32)0x000000FF)
#define WRP1_Mask                ((u32)0x0000FF00)
#define WRP2_Mask                ((u32)0x00FF0000)
#define WRP3_Mask                ((u32)0xFF000000)

/* FLASH Keys */
#define RDP_Key                  ((u16)0x00A5)
#define FLASH_KEY1               ((u32)0x45670123)
#define FLASH_KEY2               ((u32)0xCDEF89AB)

/* Delay definition */   
#define EraseTimeout             ((u32)0x00000FFF)
#define ProgramTimeout           ((u32)0x0000000F)



/* SPI SPE mask */
#define CR1_SPE_Set          ((u16)0x0040)
#define CR1_SPE_Reset        ((u16)0xFFBF)

/* SPI registers Masks */
#define CR1_CLEAR_Mask       ((u16)0x3040)



#define GET_BIT(port, pin) 	(((port)->IDR &(pin)) != 0 )
#define SET_BIT(port, pin) 	do { (port)->BSRR = (pin); } while(0)
#define CLR_BIT(port, pin) 	do { (port)->BRR = (pin); } while(0)


#if BOOTLOADER_USART == USART_1
#define USARTx USART1
#elif BOOTLOADER_USART == USART_2
#define USARTx USART2
#elif BOOTLOADER_USART == USART_3
#define USARTx USART3
#else
#error "Wrong USART configured"
#endif

/**************************************************************************************************\
\**************************************************************************************************/
void HWInit(void)
{
    /* Start with the clocks in their expected state. */
    /* Disable APB2 Peripheral Reset */
    RCC->APB2RSTR = 0x00000000;

    /* Disable APB1 Peripheral Reset */
    RCC->APB1RSTR = 0x00000000;

    /* FLITF and SRAM Clock ON */
    RCC->AHBENR = 0x00000014;

    /* Disable APB2 Peripheral Clock */
    RCC->APB2ENR = 0x00000000;

    /* Disable APB1 Peripheral Clock */
    RCC->APB1ENR = 0x00000000;

    /* Set HSION bit */
    RCC->CR |= (u32)0x00000001;

    /* Reset SW[1:0], HPRE[3:0], PPRE1[2:0], PPRE2[2:0], ADCPRE[1:0] and MCO[2:0] bits*/
    RCC->CFGR &= 0xF8FF0000;
    
    /* Reset HSEON, CSSON and PLLON bits */
    RCC->CR &= 0xFEF6FFFF;

    /* Reset HSEBYP bit */
    RCC->CR &= 0xFFFBFFFF;

    /* Reset PLLSRC, PLLXTPRE, PLLMUL[3:0] and USBPRE bits */
    RCC->CFGR &= 0xFF80FFFF;

    /* Disable all interrupts */
    RCC->CIR = 0x00000000;

	/* 2 wait states required on the flash. */
	*( ( u32 * ) 0x40022000 ) = 0x02;

	/* HCLK = SYSCLK */
    /* PCLK2 = HCLK */
    /* PCLK1 = HCLK/2 */
	/* PLLCLK = (8MHz/2) * 16 = 64 MHz. */
    RCC->CFGR |=
    	RCC_SYSCLK_Div1 |
    	(RCC_HCLK_Div1 << 3) | 
    	RCC_HCLK_Div2 |
    	RCC_PLLSource_HSI_Div2 | RCC_PLLMul_16;

	/* Enable PLL. */
    *(vu32 *) CR_PLLON_BB = (u32)ENABLE;

	/* Wait till PLL is ready. */
	while (0 == (RCC->CR & (u32) (1 << 25)))
	{
	}

	/* Select PLL as system clock source. */
	RCC->CFGR |= RCC_SYSCLKSource_PLLCLK;

	/* Wait till PLL is used as system clock source. */
	while ( ((u8)(RCC->CFGR & CFGR_SWS_Mask)) != 0x08 )
	{
	}
	
	/* Enable only USART and debug (button and LED) clocks */
	RCC->APB2ENR &= ~RCC_APB2Periph_ALL;
	RCC->APB2ENR |= RCC_APB2Periph_AFIO | RCC_APB2Periph_GPIOA | RCC_APB2Periph_GPIOB | RCC_APB2Periph_GPIOC;

	/* Enable only xyz clocks */
	RCC->APB1ENR &= ~RCC_APB1Periph_ALL;

}






void GPIO_Init(GPIO_TypeDef* GPIOx, GPIO_InitTypeDef* GPIO_InitStruct)
{
    u32 currentmode;
    u32 currentpin;
    u32 pinpos;
    u32 pos;
    u32 pinmask;
    u32 tmpreg;

/*---------------------------- GPIO Mode Configuration -----------------------*/
    currentmode = ((u32)GPIO_InitStruct->GPIO_Mode) & ((u32)0x0F);

    if ((((u32)GPIO_InitStruct->GPIO_Mode) & ((u32)0x10)) != 0x00)
    { 
      /* Output mode */
      currentmode |= (u32)GPIO_InitStruct->GPIO_Speed;
    }

/*---------------------------- GPIO CRL Configuration ------------------------*/
    /* Configure the eight low port pins */
    if (((u32)GPIO_InitStruct->GPIO_Pin & ((u32)0x00FF)) != 0x00)
    {
      tmpreg = GPIOx->CRL;

      for (pinpos = 0x00; pinpos < 0x08; pinpos++)
      {
        pos = ((u32)0x01) << pinpos;
        /* Get the port pins position */
        currentpin = (GPIO_InitStruct->GPIO_Pin) & pos;

        if (currentpin == pos)
        {
          pos = pinpos << 2;
          /* Clear the corresponding low control register bits */
          pinmask = ((u32)0x0F) << pos;
          tmpreg &= ~pinmask;

          /* Write the mode configuration in the corresponding bits */
          tmpreg |= (currentmode << pos);

          /* Reset the corresponding ODR bit */
          if (GPIO_InitStruct->GPIO_Mode == GPIO_Mode_IPD)
          {
            GPIOx->BRR = (((u32)0x01) << pinpos);
          }
          /* Set the corresponding ODR bit */
          if (GPIO_InitStruct->GPIO_Mode == GPIO_Mode_IPU)
          {
            GPIOx->BSRR = (((u32)0x01) << pinpos);
          }
        }
      }
      GPIOx->CRL = tmpreg;
    }

/*---------------------------- GPIO CRH Configuration ------------------------*/
    /* Configure the eight high port pins */
    if (GPIO_InitStruct->GPIO_Pin > 0x00FF)
    {
      tmpreg = GPIOx->CRH;
      for (pinpos = 0x00; pinpos < 0x08; pinpos++)
      {
        pos = (((u32)0x01) << (pinpos + 0x08));
        /* Get the port pins position */
        currentpin = ((GPIO_InitStruct->GPIO_Pin) & pos);
        if (currentpin == pos)
        {
          pos = pinpos << 2;
          /* Clear the corresponding high control register bits */
          pinmask = ((u32)0x0F) << pos;
          tmpreg &= ~pinmask;

          /* Write the mode configuration in the corresponding bits */
          tmpreg |= (currentmode << pos);

          /* Reset the corresponding ODR bit */
          if (GPIO_InitStruct->GPIO_Mode == GPIO_Mode_IPD)
          {
            GPIOx->BRR = (((u32)0x01) << (pinpos + 0x08));
          }
          /* Set the corresponding ODR bit */
          if (GPIO_InitStruct->GPIO_Mode == GPIO_Mode_IPU)
          {
            GPIOx->BSRR = (((u32)0x01) << (pinpos + 0x08));
          }
        }
      }
      GPIOx->CRH = tmpreg;
    }
}






/**************************************************************************************************\
\**************************************************************************************************/
void FlashUnlock(void)
{
    /* Authorize the FPEC Access */
    FLASH->KEYR = FLASH_KEY1;
    FLASH->KEYR = FLASH_KEY2;
}


/**************************************************************************************************\
\**************************************************************************************************/
void FlashLock(void)
{
    /* Set the Lock Bit to lock the FPEC and the FCR */
    FLASH->CR |= CR_LOCK_Set;
}


/*******************************************************************************
* Function Name  : FLASH_ErasePage
* Description    : Erases a specified FLASH page.
* Input          : - Page_Address: The page address to be erased.
* Output         : None
* Return         : FLASH Status: The returned value can be: FLASH_BUSY, 
*                  FLASH_ERROR_PG or FLASH_ERROR_WRP or FLASH_COMPLETE or 
*                  FLASH_TIMEOUT.
*******************************************************************************/
uint32_t FlashErasePage(uint32_t page_address)
{
    FLASH_Status status;

    /* Wait for last operation to be completed */
    status = FLASH_WaitForLastOperation(EraseTimeout);
    
    if (status == FLASH_COMPLETE)
    {
      /* if the previous operation is completed, proceed to erase the page */
      FLASH->CR|= CR_PER_Set;
      FLASH->AR = page_address; 
      FLASH->CR|= CR_STRT_Set;
      
      /* Wait for last operation to be completed */
      status = FLASH_WaitForLastOperation(EraseTimeout);

      if (status != FLASH_BUSY)
      {
        /* if the erase operation is completed, disable the PER Bit */
        FLASH->CR &= CR_PER_Reset;
      }
    }
    /* Return the Erase Status */
    return (status != FLASH_COMPLETE);
}


/*******************************************************************************
* Function Name  : FLASH_GetStatus
* Description    : Returns the FLASH Status.
* Input          : None
* Output         : None
* Return         : FLASH Status: The returned value can be: FLASH_BUSY, 
*                  FLASH_ERROR_PG or FLASH_ERROR_WRP or FLASH_COMPLETE
*******************************************************************************/
FLASH_Status FLASH_GetStatus(void)
{
    FLASH_Status flashstatus;
    
    if ((FLASH->SR & FLASH_FLAG_BSY) == FLASH_FLAG_BSY) 
    {
      flashstatus = FLASH_BUSY;
    }
    else 
    {  
      if (FLASH->SR & FLASH_FLAG_PGERR)
      { 
        flashstatus = FLASH_ERROR_PG;
      }
      else 
      {
        if (FLASH->SR & FLASH_FLAG_WRPRTERR)
        {
          flashstatus = FLASH_ERROR_WRP;
        }
        else
        {
          flashstatus = FLASH_COMPLETE;
        }
      }
    }
    /* Return the Flash Status */
    return flashstatus;
}

/*******************************************************************************
* Function Name  : FLASH_WaitForLastOperation
* Description    : Waits for a Flash operation to complete or a TIMEOUT to occur.
* Input          : - Timeout: FLASH progamming Timeout
* Output         : None
* Return         : FLASH Status: The returned value can be: FLASH_BUSY, 
*                  FLASH_ERROR_PG or FLASH_ERROR_WRP or FLASH_COMPLETE or 
*                  FLASH_TIMEOUT.
*******************************************************************************/
FLASH_Status FLASH_WaitForLastOperation(u32 Timeout)
{ 
    FLASH_Status status;
     
    /* Check for the Flash Status */
    status = FLASH_GetStatus();

    /* Wait for a Flash operation to complete or a TIMEOUT to occur */
    while ((status == FLASH_BUSY) && (Timeout != 0x00))
    {
    	MCUBusyDelay(50);
    	status = FLASH_GetStatus();
    	Timeout--;
    }

    if (Timeout == 0x00 )
    {
    	status = FLASH_TIMEOUT;
    }

    /* Return the operation status */
    return status;
}


/**************************************************************************************************\
\**************************************************************************************************/
uint32_t FlashWritePage(uint32_t page_address, const uint8_t *chunk, uint32_t chunk_len)
{
	FLASH_Status status;
	volatile uint16_t *wr_addr;
	uint16_t *data;
	uint32_t data_len;

	wr_addr = (volatile uint16_t*)page_address;
	data = (uint16_t*)chunk;
	data_len = chunk_len / 2U;
	
	/* Wait for last operation to be completed */
	status = FLASH_WaitForLastOperation(ProgramTimeout);

	if (status == FLASH_COMPLETE)
	{
	    /* if the previous operation is completed, proceed to program the new data */
	    FLASH->CR |= CR_PG_Set;
	}

	while (data_len)
	{
	    *wr_addr = *data;
	    /* Wait for last operation to be completed */
	    status = FLASH_WaitForLastOperation(ProgramTimeout);

		if (status == FLASH_COMPLETE)
		{
			if (*wr_addr != *data)
			{
				status = FLASH_ERROR_WRP;
				break;
			}
		    data++;
		    wr_addr++;
		    data_len--;
		}
		else
		{
			break;
		}
	}

	if (status != FLASH_BUSY)
	{
		/* if the program operation is completed, disable the PG Bit */
		FLASH->CR &= CR_PG_Reset;
	}

	return (status != FLASH_COMPLETE);
}

void SPI_Init(SPI_TypeDef* SPIx, SPI_InitTypeDef* SPI_InitStruct)
{
  u16 tmpreg = 0;

/*---------------------------- SPIx CR1 Configuration ------------------------*/
  /* Get the SPIx CR1 value */
  tmpreg = SPIx->CR1;
  /* Clear BIDIMode, BIDIOE, RxONLY, SSM, SSI, LSBFirst, BR, MSTR, CPOL and CPHA bits */
  tmpreg &= CR1_CLEAR_Mask;
  /* Configure SPIx: direction, NSS management, first transmitted bit, BaudRate prescaler
     master/salve mode, CPOL and CPHA */
  /* Set BIDImode, BIDIOE and RxONLY bits according to SPI_Direction value */
  /* Set SSM, SSI and MSTR bits according to SPI_Mode and SPI_NSS values */
  /* Set LSBFirst bit according to SPI_FirstBit value */
  /* Set BR bits according to SPI_BaudRatePrescaler value */
  /* Set CPOL bit according to SPI_CPOL value */
  /* Set CPHA bit according to SPI_CPHA value */
  tmpreg |= (u16)((u32)SPI_InitStruct->SPI_Direction | SPI_InitStruct->SPI_Mode |
                  SPI_InitStruct->SPI_DataSize | SPI_InitStruct->SPI_CPOL |  
                  SPI_InitStruct->SPI_CPHA | SPI_InitStruct->SPI_NSS |  
                  SPI_InitStruct->SPI_BaudRatePrescaler | SPI_InitStruct->SPI_FirstBit);
  /* Write to SPIx CR1 */
  SPIx->CR1 = tmpreg;

/*---------------------------- SPIx CRCPOLY Configuration --------------------*/
  /* Write to SPIx CRCPOLY */
  SPIx->CRCPR = SPI_InitStruct->SPI_CRCPolynomial;
}

void SPI_Cmd(SPI_TypeDef* SPIx, FunctionalState NewState)
{
  /* Check the parameters */

  if (NewState != DISABLE)
  {
    /* Enable the selected SPI peripheral */
    SPIx->CR1 |= CR1_SPE_Set;
  }
  else
  {
    /* Disable the selected SPI peripheral */
    SPIx->CR1 &= CR1_SPE_Reset;
  }
}

#if defined(OLIMEX_P103) && (OLIMEX_P103 != 0)

/**************************************************************************************************\
\**************************************************************************************************/
void LedOn(void)
{
	CLR_BIT(LED_PORT, LED_PIN);
}


/**************************************************************************************************\
\**************************************************************************************************/
void LedOff(void)
{
	SET_BIT(LED_PORT, LED_PIN);
}


/**************************************************************************************************\
\**************************************************************************************************/
uint32_t BtnState(void)
{
	return GET_BIT(BTN_PORT, BTN_PIN);
}


/**************************************************************************************************\
\**************************************************************************************************/
void BtnWaitPress(void)
{
	uint32_t debounce;
	
	debounce = 0;
  	do {
  		if (BtnState())
  		{
  			debounce++;
  		}
  		else
  		{
  			debounce = 0;
  		}
  	} while (debounce < 50000U);       
}

#endif


/**************************************************************************************************\
\**************************************************************************************************/
void MCUBusyDelay(uint32_t delay)
{
	for ( ; delay; delay--)
	{
		__asm volatile
			( 
				"	nop						\n"
				"	nop						\n"
				"	nop						\n"
				"	nop						\n"
				"	nop						\n"
			);
	}
}


