/***********************************************************************************
*	Copyright 2007 Artur Lipowski
*
***********************************************************************************/

#include "boot_config.h"

/* SRAM base address */
#define SRAM_BASE             0x20000000

/* System Control Space memory map */
#define SCS_BASE              0xE000E000
#define SCB_BASE              (SCS_BASE + 0x0D00)

/* Flash memory base address */
#define FLASH_MEM_BASE      0x8000000

#define STACK_SIZE          1024

#define APP_VEC_TAB_OFFSET  ((DEF_APP_ADDRESS - 0x8000000) & 0x1FFFFF80)


.cpu cortex-m3
.fpu softvfp   
.syntax unified
.thumb


/*****************************************************************************
 The stack gets placed into the sepatate section.
 *****************************************************************************/
.section .stack

/*****************************************************************************
 Allocate storage for the stack.
 *****************************************************************************/
StackBegin:
.space  STACK_SIZE * 4
.align 8
_initial_stack_ptr:

/*****************************************************************************
 This portion of the file goes into the text section.
 *****************************************************************************/
.section .isr_vector

/*******************************************************************************
              MCU and peripherals IRQ handlers address table
*******************************************************************************/
.word   _initial_stack_ptr
/* make sure that reset points to FLASH area */
.word   FLASH_MEM_BASE + (ResetISR - SRAM_BASE)
.word   NMIException
.word   HardFaultException
.word   MemManageException
.word   BusFaultException
.word   UsageFaultException
.word   0                       /* Reserved */
.word   0                       /* Reserved */
.word   0                       /* Reserved */
.word   0                       /* Reserved */ 
.word   FLASH_MEM_BASE + (SVCHandler - SRAM_BASE)
.word   DebugMonitor
.word   0                       /* Reserved */
.word   PendSVC
.word   SysTickHandler
.word   WWDG_IRQHandler
.word   PVD_IRQHandler
.word   TAMPER_IRQHandler
.word   RTC_IRQHandler
.word   FLASH_IRQHandler
.word   RCC_IRQHandler
.word   EXTI0_IRQHandler
.word   EXTI1_IRQHandler
.word   EXTI2_IRQHandler
.word   EXTI3_IRQHandler
.word   EXTI4_IRQHandler
.word   DMAChannel1_IRQHandler
.word   DMAChannel2_IRQHandler
.word   DMAChannel3_IRQHandler
.word   DMAChannel4_IRQHandler
.word   DMAChannel5_IRQHandler
.word   DMAChannel6_IRQHandler
.word   DMAChannel7_IRQHandler
.word   ADC_IRQHandler
.word   USB_HP_CAN1_TX_IRQHandler
.word   USB_LP_CAN1_RX0_IRQHandler
.word   CAN_RX1_IRQHandler
.word   CAN_SCE_IRQHandler
.word   EXTI9_5_IRQHandler
.word   TIM1_BRK_IRQHandler
.word   TIM1_UP_IRQHandler
.word   TIM1_TRG_COM_IRQHandler
.word   TIM1_CC_IRQHandler
.word   TIM2_IRQHandler
.word   TIM3_IRQHandler
.word   TIM4_IRQHandler
.word   I2C1_EV_IRQHandler
.word   I2C1_ER_IRQHandler
.word   I2C2_EV_IRQHandler
.word   I2C2_ER_IRQHandler
.word   SPI1_IRQHandler
.word   SPI2_IRQHandler
.word   USART1_IRQHandler
.word   USART2_IRQHandler
.word   USART3_IRQHandler
.word   EXTI15_10_IRQHandler
.word   RTCAlarm_IRQHandler
.word   USBWakeUp_IRQHandler
.word   0
.word   0
.word   0
.word   0
.word   0
.word   0
.word   0
.word   0xF108F85F              /* this is a workaround for boot in RAM mode */



.text
.pool

//*****************************************************************************
//
// The reset handler, which gets called when the processor starts.
//
//*****************************************************************************
    .globl  ResetISR
    .thumb_func
ResetISR:
    //
    // FLASH -> SRAM relocation and C runtime initialization.
    //

    //
    // Copy the code image (including .data) from flash to SRAM.
    //
    ldr     r0, =FLASH_MEM_BASE
    ldr     r1, =SRAM_BASE
    .extern _bss
    ldr     r2, =_bss
copy_loop:
        ldr     r3, [r0], #4
        str     r3, [r1], #4
        cmp     r1, r2
        blt     copy_loop

    //
    // Zero fill the .bss section.
    //
    movs    r0, #0x00000000
    .extern _ebss
    ldr     r1, =_ebss
    // r2 already contains _bss
zero_loop:
        str     r0, [r2], #4
        cmp     r2, r1
        blt     zero_loop

    //
    // Set the vector table pointer to the beginning of SRAM.
    //
    ldr     r0, =(SCB_BASE + 8)
    ldr     r1, =SRAM_BASE
    str     r1, [r0]

    //
    // "Switch" execution to the code just copied into SRAM.
    //
    ldr     r0, =main_try
    bx      r0


    //.globl  main
    .thumb_func
main_try:

    .extern HWInit
    bl      HWInit

    //
    // See if a an update should be performed.
    //
//    .extern test_user_app
//    bl      test_user_app
//    cbz     r0, CallApplication
    
    .extern boot_try
    bl      boot_try
    

    //
    // Call the application via the reset handler in its vector table.  Load
    // the address of the application vector table.
    //
    .thumb_func
CallApplication:
    //
    // Set the vector table address to the beginning of the application.
    //
    ldr     r0, =APP_VEC_TAB_OFFSET
    ldr     r1, =(SCB_BASE + 8)
    str     r0, [r1]

    //
    // Load the initial LR as it should be after the reset.
    //
    movs    lr, #0xffffffff
    
    //
    // Load the stack pointer from the application's vector table.
    //
    ldr     r0, =DEF_APP_ADDRESS
    ldr     r1, [r0]
    msr     msp, r1

    //
    // Load the initial PC from the application's vector table and branch to
    // the application's entry point.
    //
    ldr     r0, [r0, #4]
    bx      r0

#if BOOT_CALL_BY_JUMP_SUPPORT
    .thumb_func
SVCHandler:
	// load the bootloader's vector table
    ldr     r0, =0x8000000
    ldr     r1, =(SCB_BASE + 8)
    str     r0, [r1]
    //
    // Load the initial LR as it should be after the reset.
    //
    movs    lr, #0xffffffff
    
    //
    // load the stack pointer from the bootloader's vector table.
    //
    ldr     r0, =0x8000000
    ldr     r1, [r0]
    msr     msp, r1

    //
    // FLASH -> SRAM relocation and C runtime initialization.
    //

    //
    // Copy the code image (including .data) from flash to SRAM.
    //
    ldr     r0, =FLASH_MEM_BASE
    ldr     r1, =SRAM_BASE
    .extern _bss
    ldr     r2, =_bss
copy_loop2:
        ldr     r3, [r0], #4
        str     r3, [r1], #4
        cmp     r1, r2
        blt     copy_loop2

    //
    // Zero fill the .bss section.
    //
    movs    r0, #0x00000000
    .extern _ebss
    ldr     r1, =_ebss
    // r2 already contains _bss
zero_loop2:
        str     r0, [r2], #4
        cmp     r2, r1
        blt     zero_loop2

    //
    // Set the vector table pointer to the beginning of SRAM.
    //
    ldr     r0, =(SCB_BASE + 8)
    ldr     r1, =SRAM_BASE
    str     r1, [r0]

    //
    // "Switch" execution to the code just copied into SRAM.
    //
    ldr     r0, =main_force
    bx      r0
    
    //.globl  main_force
    .thumb_func
main_force:

    .extern HWInit
    bl      HWInit

    
    .extern boot_force
    bl      boot_force


    .extern CallApplication
    bl      CallApplication

#endif

.end
